/*
 *
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Swapnil V. Patil
 * Copyright (c) 2003-2004 Anand Kashyap
 * Copyright (c) 2003-2004 Gopalan Sivathanu
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>

#include "i3fs_utils.h"
#include "i3fs.h"
static const char *progname;
static char *passphrase;


// Converting policy into a bitmap 
void convert_policy_to_bitmap(char *policies, int *policy_bits, int policy_len){
	int i;
	for (i=0; i<policy_len; i++) {
		switch(policies[i]) {
			case 'D':
				*policy_bits |= DATA;
				break;
			case 'I':
				*policy_bits |= INHERIT;
				break;
			case 'p':
				*policy_bits |= PERMS;
				break;
			case 'i':
				*policy_bits |= INO;
				break;
			case 'n':
				*policy_bits |= NUM_LINKS;
				break;
			case 'u':
				*policy_bits |= UID;
				break;
			case 'g':
				*policy_bits |= GID;
				break;
			case 't':
				*policy_bits |= TYPE;
				break;
			case 's':
				*policy_bits |= SIZE;
				break;
			case 'd':
				*policy_bits |= DEV_ID;
				break;
			case 'b':
				*policy_bits |= NUM_BLOCKS;
				break;
			case 'a':
				*policy_bits |= A_TIME;
				break;
			case 'm':
				*policy_bits |= M_TIME;
				break;
			case 'c':
				*policy_bits |= C_TIME;
				break;
			default:
				break;
		}
	}
}

int add_policy_to_database(struct policy *policy_, int add_policy){
	int fd;
	struct stat buf;
	DIR *dir;
	struct dirent *dir_ent;
	struct policy new_policy;
	int retVal;
	
	fd = open(policy_->file, O_RDONLY);
    if (fd < 0) {
		perror(policy_->file);
		printf("Could not open file %s. Skipping.\n", policy_->file);
		return -1;
    }
	if (fstat(fd, &buf) == -1) {
		printf("fsat failed on file %s. Skipping. \n", policy_->file);
	} else {
		if (S_ISDIR((&buf)->st_mode) ) {
			printf ("\n%s is a directory \n", policy_->file);


			

			dir = opendir(policy_->file);
			if (!dir) {
				printf("Could not open directory %s. Skipping \n", policy_->file);
				return -1;
			}
			while(dir) {
				dir_ent = readdir(dir);
				if (!dir_ent) break;
//				printf("Name = %s\n", dir_ent->d_name);
				if (strcmp(dir_ent->d_name, ".") == 0 || strcmp(dir_ent->d_name, "..") == 0)
					continue;
				strcpy(new_policy.file, policy_->file);
				strcat(new_policy.file, "/");
				strcat(new_policy.file, dir_ent->d_name);
				new_policy.policy_bits = policy_->policy_bits;
				new_policy.freq_number = policy_->freq_number;	
				add_policy_to_database(&new_policy, add_policy);	
			}
			//return -1;
		}
	}	

	strcpy(policy_->passphrase, passphrase);


	retVal = (add_policy == 1)?ioctl(fd, ADD_POLICY, policy_):ioctl(fd, REMOVE_POLICY, policy_);

	if (retVal == -1) {
		printf ("ioctl failed on file %s. Skipping.\n", policy_->file);
	}

	printf("\nAdding/Removing policy for %s", policy_->file);

	close(fd);
	return 0;



}


// Get the policy line entry from the file and create policy entries for them into the in-kernel database and the on-disk database 
int read_policy_entry(char *policy_entry, int policy_len) 
{
	char *policies;
	struct policy *policy_;
	char *token;
	char opt_token[3];
	int fd;
	struct stat buf;
	DIR *dir;
	struct dirent *dir_ent;
	int add_policy = 0;

	policy_ = (struct policy*)malloc(sizeof(struct policy));
	policy_->policy_bits=0;

	for (token = strtok(policy_entry, " ");token; token = strtok(NULL, " ")) {
		strcpy(opt_token, token);
		//printf("\tToken = %s\n", opt_token);
		token = strtok(NULL, " ");
		if(!token) {
			printf ("Illegal argument. Skipping file %s\n", policy_->file);
			return -1;
		}
		//printf("\tNext Token = %s\n", token);
		if (!strcmp(opt_token, "-o")) {
			strcpy(policy_->file, token);
			add_policy = 1;
		} else if (!strcmp(opt_token, "-x"))
			strcpy(policy_->file, token);
		else if (!strcmp(opt_token, "-p")) {
			policies = (char *)malloc(strlen(token)+1);
			strcpy(policies, token);
			convert_policy_to_bitmap(policies, &policy_->policy_bits, strlen(token));
		} else if (!strcmp(opt_token, "-a")) {
			 if(!strcmp(token, "BLOCK"))
				policy_->policy_bits |= BLOCK;
		} else if (!strcmp(opt_token, "-g")) {
			 if(!strcmp(token, "PER_PAGE")) {
					if (!((policy_->policy_bits) & DATA)) {
						printf("Can't do per page check when policy is only for metadata. Skipping %s\n", policy_->file);
						return -1;
					}
				policy_->policy_bits |= PER_PAGE;
			}
		} else if (!strcmp(opt_token, "-c")) {
				policy_->freq_number = atoi(token);
		} else {
			printf("Illegal policy. Skipping %s\n", policy_->file);
			return -1;
		}
	}	

	//printf("Policy is %s, %d, %d\n", policy_->file, policy_->policy_bits, policy_->freq_number);
	return add_policy_to_database (policy_, add_policy);
}


void
usage(void)
{
    fprintf(stderr, "Usage: \n\t%s -f <policy_file> or \n\t%s -u [ALLOW|DISALLOW]\n", progname, progname);
    exit(1);
}

int main(int argc, char *argv[]) {
	FILE *f;
	int fpp;
	int i=0;
	char c;
	char *buf = NULL;
	int file_size, buf_size;
	struct policy *pol;
	int ret=0;

	pol = (struct policy*)malloc(sizeof(struct policy));
    
	progname = argv[0];
	if (argc != 3)
		usage();

	if (!strcmp(argv[1], "-u")) {
		passphrase = (char*)getpass("Enter the passphrase: ");
		if (!strcmp(argv[2], "ALLOW"))	{
			fpp = open(".", O_RDONLY);
			strcpy(pol->passphrase, passphrase);
			ret = ioctl(fpp, ALLOW_UPDATES, pol);
			if (!ret)	{
				printf("\nUpdates allowed\n");
				exit(0);
			}
			else	{
				printf("\nAn error has occured\n");
				perror((char*)ret);
			}
			;// ioctl for allow
		}
		else if (!strcmp(argv[2], "DISALLOW"))	{
                        fpp = open(".", O_RDONLY);
			strcpy(pol->passphrase, passphrase);
                        ioctl(fpp, DISALLOW_UPDATES, pol);

			;// ioctl for disallow
		}
		else	{
			usage();
			exit(1);
		}
	}

	if (!strcmp(argv[1], "-f")) {
	
		buf = (char *)malloc(1080);

		f = fopen(argv[2], "r");
		if (!f) {
			printf("Could not open policy file %s for reading\n", argv[2]);
			perror("Error");
			exit(1);
		}	
		passphrase = (char*)getpass("Enter the passphrase: ");
		
		while((c = fgetc(f))!= EOF) {
				if (c == '\n') {
					buf[i++] = '\0';
					read_policy_entry(buf, i);
					i=0;
					continue;
				}
				buf[i++] = c;
		}
	exit(0);
	}
	usage();	
	exit(1);
}
